"""
    CASUserFolder class allow to authenticate users via a CAS
    server and store LOCAL users created on the fly.

    CAS Server tested :
        - CAS 2.0.12 from yale
    Zope version tested :
        - 2.7.4-0
    Python version tested :
        - 2.3
"""

# Python headers
from base64 import decodestring, encodestring
from random import randint
import urllib

# Zope headers
from zLOG import LOG, DEBUG, WARNING, ERROR

from OFS.SimpleItem import Item # Basic zope object
from OFS.PropertyManager import PropertyManager # provide the 'Properties' tab with the
                                # 'manage_propertiesForm' method
from AccessControl import ClassSecurityInfo
from AccessControl.Role import RoleManager # provide the 'Ownership' tab with
                                # the 'manage_owner' method
from AccessControl.Role import DEFAULTMAXLISTUSERS
from AccessControl.ZopeSecurityPolicy import _noroles
from AccessControl.User import BasicUserFolder
from AccessControl.User import User
from AccessControl.Permissions import view_management_screens, \
                                      manage_users, \
		                      view
from Globals import DTMLFile # can use DTML files
from Globals import Persistent
from Globals import PersistentMapping
from Globals import InitializeClass
from Acquisition import Implicit


from HTMLParser import HTMLParser
class CASResponseParser(HTMLParser):
    """
    Class used to parse XML response from CAS server.
    It currently works with cas server 2.0.12 from yale.
    """
    _user = 0
    _failure = 0

    def handle_starttag(self, tag, attrs):
        # tag is returned lowercase
        if tag == 'cas:user':
	    self._user = 1
	if tag == 'cas:authenticationfailure':
	    self._failure = 1

    def handle_data(self, data):
        if self._user == 1:
            raise "user", data
	if self._failure == 1:
	    raise "failure", data

    def handle_endtag(self, tag):
            pass



class CASUserFolder(PropertyManager,
            RoleManager,
	    BasicUserFolder
	    ):
    """
    	CASUserFolder Object
    """
    
    meta_type = 'CAS User Folder'
    id        = 'acl_users'
    title     = 'CAS User Folder'

    isPrincipiaFolderish = 1
    isAUserFolder = 1
    maxlistusers = DEFAULTMAXLISTUSERS
    
    CUF_SESS_USERNAME = 'casUserName' # static variable
    cuf_default_roles = _noroles      # default roles for CAS users
    cuf_default_domains = []          # default domains for CAS users
    
    cuf_login_url     = ''
    cuf_validate_url  = ''
    cuf_logout_url    = ''
    cuf_auto_redirect = False

    security = ClassSecurityInfo()

    manage_options = (
      (
    	{'label': 'Properties', 
	 'action': 'manage_propertiesForm', 
	 'help': ('CASUserFolder', 'CASUserFolder-add.stx')
	},
	{'label': 'Contents',
	 'action': 'manage_main', 
	 'help': ('CASUserFolder', 'CASUserFolder-add.stx')
	},
      )
      + Item.manage_options            # add the 'Undo' & 'Owner' tab 
      + RoleManager.manage_options     # add the 'Security' tab
    )

    manage_propertiesForm = DTMLFile('dtml/manage_propertiesCASUserFolderForm', globals()) 
    security.declareProtected(view_management_screens, 'manage_propertiesForm')
    security.declareProtected(view_management_screens, 'manage')
    security.declareProtected(view_management_screens, 'manage_main')
    manage = manage_propertiesForm = \
        DTMLFile('dtml/manage_propertiesCASUserFolderForm', globals())

    security.declareProtected(view_management_screens, 'manage_undoForm')
    security.declareProtected(view_management_screens, 'manage_owner')
    security.declareProtected(view_management_screens, 'manage_access')
    

    # More details on properties in $ZOPE/lib/python/OFS/PropertyManager.py
    #_properties=(
    #		    { 'id':'validate_url', 
    #		      'type': 'string', 
    #		      'mode': 'w', 
    #		      'label':'Ticket validation URL' 
    #		    },
    #            )

    # ----------
    # __init__
    # ----------

    def __init__(self, login_url, validate_url, logout_url, auto_redirect=False, REQUEST=None):
	"initialise a new instance of CASUserFolder"
        Debug('__init__')
        self.id = 'acl_users'
	self.title = 'CAS User Folder'

	self.cuf_login_url     = login_url
	self.cuf_validate_url  = validate_url
	self.cuf_logout_url    = logout_url
	self.cuf_auto_redirect = auto_redirect

	# this is the container for users (BasicUserFolder naming)
	self.data = PersistentMapping()

    
    # -----------------------
    # Interface IUserFolder
    # -----------------------

    def getUser(self, username):
        """
        Returns the user object specified by name.  If there is NO
        user named 'name' in the user folder, then creates a brand
	NEW one with default values. This avoid special cases where
	folder is cleaned and user logged out.
	
	This methods means :
	    - CASUserFolder should be mentionned last source if nested into a
	      GRUF
	    - Every possible user is present => the folder will grow
	      This is however not a security hole because :
	          1) It's the authenticate method that can elect the
		     user for roles
		  2) username must be present into SESSION[CUF_SESS_USERNAME]
		     to allow this to happen from a previously validated 
		     ticket.
        """
        Debug('getUser('+username+')', '')
	if not self.data.has_key(username):
	    self._storeVolatileUser(username)
	return self.data.get(username)
	

    def getUsers(self):
        """
        Returns a sorted sequence of all user objects which reside in the user
        folder.
        """
        Debug('getUsers()', '')
	data = self.data
	names = data.keys()
	names.sort()
	return [data[n] for n in names]
	

    def getUserNames(self):
        """
        Returns a sorted sequence of names of the users which have connected
        recently.
	This list is volatile and can be emptyed when this object is reloaded 
	from ZODB.
        """
        Debug('getUserNames()', '')
	names = self.data.keys()
	names.sort()
	return names
	

    # --------------------------
    # BasicUserFolder Overrides
    # --------------------------

    def getUserById(self, id, default=None):
	return self.getUser(id)
    

    def hasUsers(self):
        Debug('hasUsers()', '')
    	return True # always true => we depend on external data

    #
    # def validate(self, request, auth='', roles=_noroles):
    #     # the validate() method is left to parent class
    #


    # --------------------------------------------
    # BasicUserFolder "_do" method implementation 
    # --------------------------------------------
    
    def _doChangeUser(self, name, password, roles, domains, **kw):
        """
        This method is needed by GRUF.
        Roles are the way GRUF affect groups on users.
        """
        user=self.data[name]
        if password is not None:
            user.__=password
        user.roles=roles
        user.domains=domains

    def _doAddUser(self, name, password, roles, domains, **kw):
        """
	Add a new user or overwrite an existing one, we don't care!
	These users are temporary and used only as markers.
	To avoid hidden security hole however, a random pwd is generated
	"""
	random_pwd = encodestring(str(randint(1234, 999999999))+str(randint(4321, 999999999)))
        self.data[name] = User(name,random_pwd,roles,domains)

    def _doDelUsers(self, names):
        for name in names:
	    del self.data[name]
	
    
    
	
    # -------------------------------------
    # BasicUserFolder Authentication stuff
    # -------------------------------------

    def identify(self, auth):
        """
	Returns: 'CAS', 'CAS'
	         These values are not used at all, but they are mandatory
		 if we are in a GroupUserFolder (auth-patched) for authenticate
		 to be called.
        """
        Debug('identify(auth=%s)'%(str(auth)), '')
	return 'CAS', 'CAS'
	

    def authenticate(self, name, password, REQUEST):
        """
	We try here to identify a CAS user by checking its ticket against the
	CAS server.
        If a name and password are provided we just return, as it is
	no more a cas authentication.

	Returns: User instance
	"""
        Debug('authenticate(name=%s, password=%s, %s)'%(str(name), 'XXXXX', 'REQUEST'), 
	      'SESSION=%s'%(REQUEST.SESSION))
	
	# has the browser sent a CAS ticket?
	ticket = REQUEST.form.get('ticket')

        if ticket is None:
	    # no ticket, have we a username in the session?
	    if REQUEST.SESSION.has_key(self.CUF_SESS_USERNAME):
	        username = REQUEST.SESSION[self.CUF_SESS_USERNAME]
	        return self.getUser(username)
	    Debug('authenticate(%s, %s, %s) => %s'%(str(name), 'XXXXX', 'REQUEST', 'CACHE MISS'), '')
	    if self.cuf_auto_redirect:
	        REQUEST.RESPONSE.redirect('%s?service=%s'%(self.cuf_login_url, self._getService(REQUEST)))
	else:
	    # HERE we talk to CAS server
	    username = self._validateTicket(ticket, REQUEST)
	    Debug('authenticate(%s, %s, %s) => %s'%(str(name), 'XXXXX', 'REQUEST', username), '')
	    if username is not None:
	        # CAS TICKET VALIDATED
	        # register validated user into the session
	        REQUEST.SESSION[self.CUF_SESS_USERNAME] = username
	        self._storeVolatileUser(username, REQUEST)

	        # RETURN the user previously stored (or refreshed)
		return self.getUser(username)
	    else:
	        # TICKET NOT VALIDATED
		# this can happen if a user use an old bookmark for example.
		# We try then to fallaback to a previous session
		if REQUEST.SESSION.has_key(self.CUF_SESS_USERNAME):
		    username = REQUEST.SESSION[self.CUF_SESS_USERNAME]
		    return self.getUser(username)
	
	# defer the authenticate() to other User Folder if any
	return None


    #
    # def authorize(self, user, accessed, container, name, value, roles):
    # => the authorize() method is left to parent class
    #



    # ------------------------------
    # CAS User Folder PRIVATE stuff
    # ------------------------------
    
    security.declarePrivate('_storeVolatileUser')
    def _storeVolatileUser(self, username, REQUEST=None):
        """
	store a "ticket granted" user or refresh an older one
	"""
        Debug('_storeVolatileUser(%s, REQUEST)'%(username), '')
        user = self.data.get(username)
	if user is None:
	    self._doAddUser(username, 'DUMMY', self.cuf_default_roles, self.cuf_default_domains)

    security.declarePrivate('_getService')
    def _getService(self, REQUEST):
        """
	Extract URL and QUERY args from request to provide a valid CAS
	service URL.
	"""
	return urllib.quote(REQUEST['URL'])
	
	## standby on this code -- this is an attempt on passing the GET args
	## in the CAS service string. Don't work fine in Plone.
	#
	#query_string = REQUEST['QUERY_STRING']
	#Debug('_getService(%s,%s) => '%(REQUEST['URL'],query_string), )
	#if query_string is None or query_string == "":
        #    svc = urllib.quote(REQUEST['URL'])
	#else:
	#    svc = REQUEST['URL']
	#    Q2=""
	#    list = query_string.split('&')
	#    if len(list) > 0:
	#	flag=0
      	#	for var in list:
        #  	    if var.find('ticket=') == 0:
        #      	        continue
        #  	    if flag:
        #      		Q2=Q2+'&'+var
        #  	    else:
        #      		Q2=Q2+'?'+var
        #      		flag=1
        #
      	#	if Q2 != "":
        #	    svc=svc+Q2
        #
	#        svc = urllib.quote( svc )
        #	
	#Debug('_getService(REQUEST) => ', svc)
	#return svc
    
    security.declarePrivate('_validateTicket')
    def _validateTicket(self, ticket, REQUEST):
        '''
	Validate the ticket we got in the GET data to the CAS server
	RETURNS: username or None
	'''
        Debug('_validateTicket('+ticket+', REQUEST)', '')
	
        # prepare the GET parameters for checking the login
	service = self._getService(REQUEST)
        checkparams = "?service=%s&ticket=%s" % (service, ticket)
        # check the ticket
	try:
            casdata = urllib.URLopener().open(self.cuf_validate_url + checkparams)
	except:
	    Error('_validateTicket(REQUEST[URL]=%s, ticket=%s)'%(REQUEST['URL'],ticket), 'CANNOT CONTACT CAS SERVER')
	    # The end user will got an 'IOError' which is probably best than leaving
	    # him silently unautenticated.
	    raise
        test = casdata.readline().strip()
	Debug('_validateTicket()', 'CAS RESPONSE 1st line: ' + str(test))
        if test == 'yes':
            # user is validated (CAS architecture 1.0)
            username = casdata.readline().strip()
            return username
        else:
	    # If we are here, either the user is NOT validated, either
	    # it may be an XML response (CAS architecture 2.0).
	    # Then try to parse an XML response
	    if test.find("cas:serviceResponse") > 0:
	        parser = CASResponseParser()
		try:
		    while test != "":
		        parser.feed(test.strip())
		        test = casdata.readline()
	        except "user", user:
		    return user.strip()
	        except "failure", failure_msg:
        	    Warning('_validateTicket('+REQUEST['URL']+', '+ticket+') ', "FAILURE => %s"%(failure_msg.strip()))
		except:
		    # hummm, maybe it was not XML...
		    pass
		
            # some unknown authentication error occurred
            return None
 	

    # ------------
    # Management 
    # ------------
       
    # The method used to edit this object
    def manage_propertiesAction(self, login_url, validate_url, logout_url, auto_redirect="", RESPONSE=None):
        "Changes the instance values"
	self.cuf_login_url     = login_url
        self.cuf_validate_url  = validate_url
	self.cuf_logout_url    = logout_url
	self.cuf_auto_redirect = not not auto_redirect
        self._p_changed = 1
        RESPONSE.redirect('manage_propertiesForm')

    #
    # END of CASUserFolder Class
    #					    

#
# LOG
#
def Debug(summary, detail=""):
    pass
    #LOG("CASUserFolder", DEBUG, summary, detail="")

def Error(summary, detail=""):
    LOG("CASUserFolder", ERROR, summary, detail="")

def Warning(summary, detail=""):
    LOG("CASUserFolder", WARNING, summary, detail="")


#
# Product Administration
#

def manage_addCASUserFolderAction(self, 
				  login_url,
			          validate_url,
				  logout_url,
				  auto_redirect="",
				  REQUEST=None):
   "Add a CASUserFolder instance to a folder."
   auto_redirect = not not auto_redirect # magic not not feature!
   Debug("login_url=%s,validate_url=%s,logout_url=%s,auto_redirect=%s"%(login_url,validate_url,logout_url,auto_redirect))
   self._setObject('acl_users', CASUserFolder(login_url,validate_url,logout_url,auto_redirect))
   if REQUEST is not None:
        return self.manage_main(self, REQUEST)

# The form used to get the instance id from the user.
manage_addCASUserFolderForm = DTMLFile('dtml/manage_addCASUserFolderForm', globals())


#
# Initialization
#
InitializeClass(CASUserFolder)


